import datetime
import hashlib
from typing import Any, Literal, Optional

import requests

from dsp.modules.lm import LM


def post_request_metadata(model_name, prompt):
    """Creates a serialized request object for the Ollama API."""
    timestamp = datetime.datetime.now().timestamp()
    id_string = str(timestamp) + model_name + prompt
    hashlib.sha1().update(id_string.encode("utf-8"))
    id_hash = hashlib.sha1().hexdigest()
    return {"id": f"chatcmpl-{id_hash}", "object": "chat.completion", "created": int(timestamp), "model": model_name}


class OllamaLocal(LM):
    """Wrapper around a locally hosted Ollama model (API: https://github.com/jmorganca/ollama/blob/main/docs/modelfile.md#valid-parameters-and-values and https://github.com/jmorganca/ollama/blob/main/docs/api.md#generate-a-completion).
    Returns dictionary info in the OpenAI API style (https://platform.openai.com/docs/api-reference/chat/object).

    Args:
        model (str, optional): Name of Ollama model. Defaults to "llama2".
        model_type (Literal["chat", "text"], optional): The type of model that was specified. Mainly to decide the optimal prompting strategy. Defaults to "text".
        base_url (str):  Protocol, host name, and port to the served ollama model. Defaults to "http://localhost:11434" as in ollama docs.
        timeout_s (float): Timeout period (in seconds) for the post request to llm.
        format (str): The format to return a response in. Currently the only accepted value is `json`
        system (str): System Prompt to use when running in `text` mode.
        **kwargs: Additional arguments to pass to the API.
    """

    def __init__(
        self,
        model: str = "llama2",
        model_type: Literal["chat", "text"] = "text",
        base_url: str = "http://localhost:11434",
        timeout_s: float = 120,
        temperature: float = 0.0,
        max_tokens: int = 150,
        top_p: int = 1,
        top_k: int = 20,
        frequency_penalty: float = 0,
        presence_penalty: float = 0,
        n: int = 1,
        num_ctx: int = 1024,
        format: Optional[Literal["json"]] = None,
        system: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(model)

        self.provider = "ollama"
        self.model_type = model_type
        self.base_url = base_url
        self.model_name = model
        self.timeout_s = timeout_s
        self.format = format
        self.system = system

        self.kwargs = {
            "temperature": temperature,
            "max_tokens": max_tokens,
            "top_p": top_p,
            "top_k": top_k,
            "frequency_penalty": frequency_penalty,
            "presence_penalty": presence_penalty,
            "n": n,
            "num_ctx": num_ctx,
            **kwargs,
        }

        # Ollama uses num_predict instead of max_tokens
        self.kwargs["num_predict"] = self.kwargs["max_tokens"]

        self.history: list[dict[str, Any]] = []
        self.version = kwargs["version"] if "version" in kwargs else ""

        # Ollama occasionally does not send `prompt_eval_count` in response body.
        # https://github.com/stanfordnlp/dspy/issues/293
        self._prev_prompt_eval_count = 0

    def basic_request(self, prompt: str, **kwargs):
        raw_kwargs = kwargs

        kwargs = {**self.kwargs, **kwargs}

        request_info = post_request_metadata(self.model_name, prompt)
        request_info["choices"] = []
        settings_dict = {
            "model": self.model_name,
            "options": {k: v for k, v in kwargs.items() if k not in ["n", "max_tokens"]},
            "stream": False,
        }

        # Set the format if it was defined
        if self.format:
            settings_dict["format"] = self.format

        if self.model_type == "chat":
            settings_dict["messages"] = [{"role": "user", "content": prompt}]
        else:
            # Overwrite system prompt defined in modelfile
            if self.system:
                settings_dict["system"] = self.system

            settings_dict["prompt"] = prompt

        urlstr = f"{self.base_url}/api/chat" if self.model_type == "chat" else f"{self.base_url}/api/generate"
        tot_eval_tokens = 0
        for i in range(kwargs["n"]):
            response = requests.post(urlstr, json=settings_dict, timeout=self.timeout_s)

            # Check if the request was successful (HTTP status code 200)
            if response.status_code != 200:
                # If the request was not successful, print an error message
                print(f"Error: CODE {response.status_code} - {response.text}")

            response_json = response.json()

            text = (
                response_json.get("message").get("content")
                if self.model_type == "chat"
                else response_json.get("response")
            )
            request_info["choices"].append(
                {
                    "index": i,
                    "message": {
                        "role": "assistant",
                        "content": "".join(text),
                    },
                    "finish_reason": "stop",
                },
            )
            tot_eval_tokens += response_json.get("eval_count", self._prev_prompt_eval_count)
        request_info["additional_kwargs"] = {k: v for k, v in response_json.items() if k not in ["response"]}

        request_info["usage"] = {
            "prompt_tokens": response_json.get("prompt_eval_count", self._prev_prompt_eval_count),
            "completion_tokens": tot_eval_tokens,
            "total_tokens": response_json.get("prompt_eval_count", self._prev_prompt_eval_count) + tot_eval_tokens,
        }

        history = {
            "prompt": prompt,
            "response": request_info,
            "kwargs": kwargs,
            "raw_kwargs": raw_kwargs,
        }
        self.history.append(history)

        return request_info

    def request(self, prompt: str, **kwargs):
        """Wrapper for requesting completions from the Ollama model."""
        if "model_type" in kwargs:
            del kwargs["model_type"]

        return self.basic_request(prompt, **kwargs)

    def _get_choice_text(self, choice: dict[str, Any]) -> str:
        return choice["message"]["content"]

    def __call__(
        self,
        prompt: str,
        only_completed: bool = True,
        return_sorted: bool = False,
        **kwargs,
    ) -> list[dict[str, Any]]:
        """Retrieves completions from Ollama.

        Args:
            prompt (str): prompt to send to Ollama
            only_completed (bool, optional): return only completed responses and ignores completion due to length. Defaults to True.
            return_sorted (bool, optional): sort the completion choices using the returned probabilities. Defaults to False.

        Returns:
            list[dict[str, Any]]: list of completion choices
        """

        assert only_completed, "for now"
        assert return_sorted is False, "for now"

        response = self.request(prompt, **kwargs)

        choices = response["choices"]

        completed_choices = [c for c in choices if c["finish_reason"] != "length"]

        if only_completed and len(completed_choices):
            choices = completed_choices

        completions = [self._get_choice_text(c) for c in choices]

        return completions

    def copy(self, **kwargs):
        """Returns a copy of the language model with the same parameters."""
        kwargs = {**self.kwargs, **kwargs}

        return self.__class__(
            model=self.model_name,
            model_type=self.model_type,
            base_url=self.base_url,
            timeout_s=self.timeout_s,
            **kwargs,
        )


import inspect
import uuid
from typing import Any

import dsp
import dspy

#################### Assertion Helpers ####################


def _build_error_msg(feedback_msgs):
    """Build an error message from a list of feedback messages."""
    return "\n".join([msg for msg in feedback_msgs])


#################### Assertion Exceptions ####################


class DSPyAssertionError(AssertionError):
    """Custom exception raised when a DSPy `Assert` fails."""

    def __init__(
        self,
        id: str,
        msg: str,
        target_module: Any = None,
        state: Any = None,
        is_metric: bool = False,
    ) -> None:
        super().__init__(msg)
        self.id = id
        self.msg = msg
        self.target_module = target_module
        self.state = state
        self.is_metric = is_metric


class DSPySuggestionError(AssertionError):
    """Custom exception raised when a DSPy `Suggest` fails."""

    def __init__(
        self,
        id: str,
        msg: str,
        target_module: Any = None,
        state: Any = None,
        is_metric: bool = False,
    ) -> None:
        super().__init__(msg)
        self.id = id
        self.msg = msg
        self.target_module = target_module
        self.state = state
        self.is_metric = is_metric


#################### Assertion Primitives ####################


class Constraint:
    def __init__(
        self,
        result: bool,
        msg: str = "",
        target_module=None,
        is_metric: bool = False,
    ):
        self.id = str(uuid.uuid4())
        self.result = result
        self.msg = msg
        self.target_module = target_module
        self.is_metric = is_metric

        self.__call__()


class Assert(Constraint):
    """DSPy Assertion"""

    def __call__(self) -> bool:
        if isinstance(self.result, bool):
            if self.result:
                return True
            elif dspy.settings.bypass_assert:
                dspy.logger.error(f"AssertionError: {self.msg}")
                return True
            else:
                dspy.logger.error(f"AssertionError: {self.msg}")
                raise DSPyAssertionError(
                    id=self.id,
                    msg=self.msg,
                    target_module=self.target_module,
                    state=dsp.settings.trace,
                    is_metric=self.is_metric,
                )
        else:
            raise ValueError("Assertion function should always return [bool]")


class Suggest(Constraint):
    """DSPy Suggestion"""

    def __call__(self) -> Any:
        if isinstance(self.result, bool):
            if self.result:
                return True
            elif dspy.settings.bypass_suggest:
                dspy.logger.info(f"SuggestionFailed: {self.msg}")
                return True
            else:
                dspy.logger.info(f"SuggestionFailed: {self.msg}")
                raise DSPySuggestionError(
                    id=self.id,
                    msg=self.msg,
                    target_module=self.target_module,
                    state=dsp.settings.trace,
                    is_metric=self.is_metric,
                )
        else:
            raise ValueError("Suggestion function should always return [bool]")


#################### Assertion Handlers ####################


def noop_handler(func):
    """Handler to bypass assertions and suggestions.

    Now both assertions and suggestions will become noops.
    """

    def wrapper(*args, **kwargs):
        with dspy.settings.context(bypass_assert=True, bypass_suggest=True):
            return func(*args, **kwargs)

    return wrapper


def bypass_suggest_handler(func):
    """Handler to bypass suggest only.

    If a suggestion fails, it will be logged but not raised.
    And If an assertion fails, it will be raised.
    """

    def wrapper(*args, **kwargs):
        with dspy.settings.context(bypass_suggest=True, bypass_assert=False):
            return func(*args, **kwargs)

    return wrapper


def bypass_assert_handler(func):
    """Handler to bypass assertion only.

    If a assertion fails, it will be logged but not raised.
    And If an assertion fails, it will be raised.
    """

    def wrapper(*args, **kwargs):
        with dspy.settings.context(bypass_assert=True):
            return func(*args, **kwargs)

    return wrapper


def assert_no_except_handler(func):
    """Handler to ignore assertion failure and return None."""

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except DSPyAssertionError:
            return None

    return wrapper


def backtrack_handler(func, bypass_suggest=True, max_backtracks=2):
    """Handler for backtracking suggestion and assertion.

    Re-run the latest predictor up to `max_backtracks` times,
    with updated signature if an assertion fails. updated signature adds a new
    input field to the signature, which is the feedback.
    """

    def wrapper(*args, **kwargs):
        error_msg, result = None, None
        with dspy.settings.lock:
            dspy.settings.backtrack_to = None
            dspy.settings.suggest_failures = 0
            dspy.settings.assert_failures = 0

            # Predictor -> List[feedback_msg]
            dspy.settings.predictor_feedbacks = {}

            current_error = None
            for i in range(max_backtracks + 1):
                if i > 0 and dspy.settings.backtrack_to is not None:
                    # generate values for new fields
                    feedback_msg = _build_error_msg(
                        dspy.settings.predictor_feedbacks[dspy.settings.backtrack_to],
                    )

                    dspy.settings.backtrack_to_args = {
                        "feedback": feedback_msg,
                        "past_outputs": past_outputs,
                    }

                # if last backtrack: ignore suggestion errors
                if i == max_backtracks:
                    if isinstance(current_error, DSPyAssertionError):
                        raise current_error
                    dsp.settings.trace.clear()
                    result = bypass_suggest_handler(func)(*args, **kwargs) if bypass_suggest else None
                    break
                else:
                    try:
                        dsp.settings.trace.clear()
                        result = func(*args, **kwargs)
                        break
                    except (DSPySuggestionError, DSPyAssertionError) as e:
                        if not current_error:
                            current_error = e
                        error_id, error_msg, error_target_module, error_state = (
                            e.id,
                            e.msg,
                            e.target_module,
                            e.state[-1],
                        )

                        # increment failure count depending on type of error
                        if isinstance(e, DSPySuggestionError) and e.is_metric:
                            dspy.settings.suggest_failures += 1
                        elif isinstance(e, DSPyAssertionError) and e.is_metric:
                            dspy.settings.assert_failures += 1

                        if dsp.settings.trace:
                            if error_target_module:
                                for i in range(len(dsp.settings.trace) - 1, -1, -1):
                                    trace_element = dsp.settings.trace[i]
                                    mod = trace_element[0]
                                    if mod == error_target_module:
                                        error_state = e.state[i]
                                        dspy.settings.backtrack_to = mod
                                        break
                            else:
                                dspy.settings.backtrack_to = dsp.settings.trace[-1][0]

                            if dspy.settings.backtrack_to is None:
                                dspy.logger.error("Module not found in trace. If passing a DSPy Signature, please specify the intended module for the assertion (e.g., use `target_module = self.my_module(my_signature)` instead of `target_module =  my_signature`).")

                            # save unique feedback message for predictor
                            if error_msg not in dspy.settings.predictor_feedbacks.setdefault(
                                dspy.settings.backtrack_to,
                                [],
                            ):
                                dspy.settings.predictor_feedbacks[dspy.settings.backtrack_to].append(error_msg)

                            # use `new_signature` if available (CoT)
                            if hasattr(error_state[0], 'new_signature'):
                                output_fields = error_state[0].new_signature.output_fields
                            else:
                                output_fields = error_state[0].signature.output_fields
                            past_outputs = {}
                            for field_name in output_fields.keys():
                                past_outputs[field_name] = getattr(
                                    error_state[2],
                                    field_name,
                                    None,
                                )

                            # save latest failure trace for predictor per suggestion
                            error_ip = error_state[1]
                            error_op = error_state[2].__dict__["_store"]
                            error_op.pop("_assert_feedback", None)
                            error_op.pop("_assert_traces", None)

                        else:
                            dspy.logger.error(
                                "UNREACHABLE: No trace available, this should not happen. Is this run time?",
                            )

            return result

    return wrapper


def handle_assert_forward(assertion_handler, **handler_args):
    def forward(self, *args, **kwargs):
        args_to_vals = inspect.getcallargs(self._forward, *args, **kwargs)

        # if user has specified a bypass_assert flag, set it
        if "bypass_assert" in args_to_vals:
            dspy.settings.configure(bypass_assert=args_to_vals["bypass_assert"])

        wrapped_forward = assertion_handler(self._forward, **handler_args)
        return wrapped_forward(*args, **kwargs)

    return forward


default_assertion_handler = backtrack_handler


def assert_transform_module(
    module,
    assertion_handler=default_assertion_handler,
    **handler_args,
):
    """
    Transform a module to handle assertions.
    """
    if not getattr(module, "forward", False):
        raise ValueError(
            "Module must have a forward method to have assertions handled.",
        )
    if getattr(module, "_forward", False):
        dspy.logger.info(
            f"Module {module.__class__.__name__} already has a _forward method. Skipping...",
        )
        pass  # TODO warning: might be overwriting a previous _forward method

    module._forward = module.forward
    module.forward = handle_assert_forward(assertion_handler, **handler_args).__get__(
        module,
    )

    if all(
        map(lambda p: isinstance(p[1], dspy.retry.Retry), module.named_predictors()),
    ):
        pass  # we already applied the Retry mapping outside
    elif all(
        map(lambda p: not isinstance(p[1], dspy.retry.Retry), module.named_predictors()),
    ):
        module.map_named_predictors(dspy.retry.Retry)
    else:
        raise RuntimeError("Module has mixed predictors, can't apply Retry mapping.")

    module._assert_transformed = True

    return module



# # To Run:
# # python -m dsp.modules.hf_server --port 4242 --model "google/flan-t5-base"

# # To Query:
# # curl -d '{"prompt":".."}' -X POST "http://0.0.0.0:4242" -H 'Content-Type: application/json'
# # Or use the HF client. TODO: Add support for kwargs to the server.


# from functools import lru_cache
# import argparse
# import time
# import random
# import os
# import sys
# import uvicorn
# import warnings

# from fastapi import FastAPI
# from pydantic import BaseModel
# from argparse import ArgumentParser
# from starlette.middleware.cors import CORSMiddleware

# from dsp.modules.hf import HFModel


# class Query(BaseModel):
#     prompt: str
#     kwargs: dict = {}


# warnings.filterwarnings("ignore")

# app = FastAPI()
# app.add_middleware(
#     CORSMiddleware, allow_origins=["*"], allow_methods=["*"], allow_headers=["*"]
# )

# parser = argparse.ArgumentParser("Server for Hugging Face models")
# parser.add_argument("--port", type=int, required=True, help="Server port")
# parser.add_argument("--model", type=str, required=True, help="Hugging Face model")
# args = parser.parse_args()
# # TODO: Convert this to a log message
# print(f"#> Loading the language model {args.model}")
# lm = HFModel(args.model)


# @lru_cache(maxsize=None)
# def generate(prompt, **kwargs):
#     global lm
#     generateStart = time.time()
#     # TODO: Convert this to a log message
#     print(f'#> kwargs: "{kwargs}" (type={type(kwargs)})')
#     response = lm._generate(prompt, **kwargs)
#     # TODO: Convert this to a log message
#     print(f'#> Response: "{response}"')
#     latency = (time.time() - generateStart) * 1000.0
#     response["latency"] = latency
#     print(f'#> Latency:', '{:.3f}'.format(latency / 1000.0), 'seconds')
#     return response


# @app.post("/")
# async def generate_post(query: Query):
#     return generate(query.prompt, **query.kwargs)


# if __name__ == "__main__":
#     uvicorn.run(
#         app,
#         host="0.0.0.0",
#         port=args.port,
#         reload=False,
#         log_level="info",
#     )  # can make reload=True later



import logging
import os
from typing import Any, Optional

import backoff
from dsp.utils.settings import settings

from dsp.modules.lm import LM

try:
    import anthropic

    anthropic_rate_limit = anthropic.RateLimitError
except ImportError:
    anthropic_rate_limit = Exception

logger = logging.getLogger(__name__)
BASE_URL = "https://api.anthropic.com/v1/messages"


def backoff_hdlr(details):
    """Handler from https://pypi.org/project/backoff/."""
    print(
        "Backing off {wait:0.1f} seconds after {tries} tries "
        "calling function {target} with kwargs "
        "{kwargs}".format(**details),
    )


def giveup_hdlr(details):
    """Wrapper function that decides when to give up on retry."""
    if "rate limits" in details.message:
        return False
    return True


class Claude(LM):
    """Wrapper around anthropic's API. Supports both the Anthropic and Azure APIs."""

    def __init__(
        self,
        model: str = "claude-3-opus-20240229",
        api_key: Optional[str] = None,
        api_base: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(model)
        try:
            from anthropic import Anthropic
        except ImportError as err:
            raise ImportError("Claude requires `pip install anthropic`.") from err

        self.provider = "anthropic"
        self.api_key = api_key = os.environ.get("ANTHROPIC_API_KEY") if api_key is None else api_key
        self.api_base = BASE_URL if api_base is None else api_base
        self.kwargs = {
            "temperature": kwargs.get("temperature", 0.0),
            "max_tokens": min(kwargs.get("max_tokens", 4096), 4096),
            "top_p": kwargs.get("top_p", 1.0),
            "top_k": kwargs.get("top_k", 1),
            "n": kwargs.pop("n", kwargs.pop("num_generations", 1)),
            **kwargs,
        }
        self.kwargs["model"] = model
        self.history: list[dict[str, Any]] = []
        self.client = Anthropic(api_key=api_key, base_url=api_base)

    def log_usage(self, response):
        """Log the total tokens from the Anthropic API response."""
        usage_data = response.usage
        if usage_data:
            total_tokens = usage_data.input_tokens + usage_data.output_tokens
            logger.debug(f"Anthropic Total Token Response Usage: {total_tokens}")

    def basic_request(self, prompt: str, **kwargs):
        raw_kwargs = kwargs
        kwargs = {**self.kwargs, **kwargs}
        # caching mechanism requires hashable kwargs
        kwargs["messages"] = [{"role": "user", "content": prompt}]
        kwargs.pop("n")
        response = self.client.messages.create(**kwargs)
        history = {
            "prompt": prompt,
            "response": response,
            "kwargs": kwargs,
            "raw_kwargs": raw_kwargs,
        }
        self.history.append(history)
        return response

    @backoff.on_exception(
        backoff.expo,
        (anthropic_rate_limit),
        max_time=settings.backoff_time,
        max_tries=8,
        on_backoff=backoff_hdlr,
        giveup=giveup_hdlr,
    )
    def request(self, prompt: str, **kwargs):
        """Handles retrieval of completions from Anthropic whilst handling API errors."""
        return self.basic_request(prompt, **kwargs)

    def __call__(self, prompt, only_completed=True, return_sorted=False, **kwargs):
        """Retrieves completions from Anthropic.

        Args:
            prompt (str): prompt to send to Anthropic
            only_completed (bool, optional): return only completed responses and ignores completion due to length. Defaults to True.
            return_sorted (bool, optional): sort the completion choices using the returned probabilities. Defaults to False.

        Returns:
            list[str]: list of completion choices
        """
        assert only_completed, "for now"
        assert return_sorted is False, "for now"
        # per eg here: https://docs.anthropic.com/claude/reference/messages-examples
        # max tokens can be used as a proxy to return smaller responses
        # so this cannot be a proper indicator for incomplete response unless it isnt the user-intent.
        n = kwargs.pop("n", 1)
        completions = []
        for _ in range(n):
            response = self.request(prompt, **kwargs)
            self.log_usage(response)
            if only_completed and response.stop_reason == "max_tokens":
                continue
            completions = [c.text for c in response.content]
        return completions



import dsp
import dspy

from .predict import Predict

# TODO: FIXME: Insert this right before the *first* output field. Also rewrite this to use the new signature system.

class ChainOfThoughtWithHint(Predict):
    def __init__(self, signature, rationale_type=None, activated=True, **config):
        super().__init__(signature, **config)
        self.activated = activated
        signature = self.signature

        *keys, last_key = signature.fields.keys()
        rationale_type = rationale_type or dspy.OutputField(
            prefix="Reasoning: Let's think step by step in order to",
            desc="${produce the " + last_key + "}. We ...",
        )
        self.extended_signature1 = self.signature.insert(-2, "rationale", rationale_type, type_=str)

        DEFAULT_HINT_TYPE = dspy.OutputField()
        self.extended_signature2 = self.extended_signature1.insert(-2, "hint", DEFAULT_HINT_TYPE, type_=str)
    
    def forward(self, **kwargs):
        signature = self.signature

        if self.activated is True or (self.activated is None and isinstance(dsp.settings.lm, dsp.GPT3)):
            if 'hint' in kwargs and kwargs['hint']:
                signature = self.extended_signature2
            else:
                signature = self.extended_signature1
        
        return super().forward(signature=signature, **kwargs)


"""
TODO: In principle, we can update the field's prefix during forward too to fill any thing based on the input args.

IF the user didn't overwrite our default rationale_type.
"""


import dspy


class SemanticRecallPrecision(dspy.Signature):
    """
    Compare a system's response to the ground truth to compute its recall and precision.
    If asked to reason, enumerate key ideas in each response, and whether they are present in the other response.
    """

    question: str = dspy.InputField()
    ground_truth: str = dspy.InputField()
    system_response: str = dspy.InputField()
    recall: float = dspy.OutputField(desc="fraction (out of 1.0) of ground truth covered by the system response")
    precision: float = dspy.OutputField(desc="fraction (out of 1.0) of system response covered by the ground truth")


def f1_score(precision, recall):
    return 0.0 if precision + recall == 0 else 2 * (precision * recall) / (precision + recall)


class SemanticF1(dspy.Module):
    def __init__(self, threshold=0.66):
        self.threshold = threshold
        self.module = dspy.ChainOfThought(SemanticRecallPrecision)

    def forward(self, example, pred, trace=None):
        scores = self.module(question=example.question, ground_truth=example.response, system_response=pred.response)
        score = f1_score(scores.precision, scores.recall)

        return score if trace is None else score >= self.threshold


"""
Soon-to-be deprecated Signatures & Modules Below.
"""


class AnswerCorrectnessSignature(dspy.Signature):
    """Verify that the predicted answer matches the gold answer."""

    question = dspy.InputField()
    gold_answer = dspy.InputField(desc="correct answer for question")
    predicted_answer = dspy.InputField(desc="predicted answer for question")
    is_correct = dspy.OutputField(desc="True or False")


class AnswerCorrectness(dspy.Module):
    def __init__(self):
        super().__init__()
        self.evaluate_correctness = dspy.ChainOfThought(AnswerCorrectnessSignature)

    def forward(self, question, gold_answer, predicted_answer):
        return self.evaluate_correctness(question=question, gold_answer=gold_answer, predicted_answer=predicted_answer)


class AnswerFaithfulnessSignature(dspy.Signature):
    """Verify that the predicted answer is based on the provided context."""

    context = dspy.InputField(desc="relevant facts for producing answer")
    question = dspy.InputField()
    answer = dspy.InputField(desc="often between 1 and 5 words")
    is_faithful = dspy.OutputField(desc="True or False")


class AnswerFaithfulness(dspy.Module):
    def __init__(self):
        super().__init__()
        self.evaluate_faithfulness = dspy.ChainOfThought(AnswerFaithfulnessSignature)

    def forward(self, context, question, answer):
        return self.evaluate_faithfulness(context=context, question=question, answer=answer)



from typing import Any, Dict, List, Optional, Type, Union

from dspy.clients.anyscale import FinetuneJobAnyScale, finetune_anyscale
from dspy.clients.finetune import FinetuneJob, TrainingMethod
from dspy.clients.openai import FinetuneJobOpenAI, finetune_openai
from dspy.utils.logging import logger

_PROVIDER_ANYSCALE = "anyscale"
_PROVIDER_OPENAI = "openai"


def get_provider_finetune_job_class(provider: str) -> Type[FinetuneJob]:
    """Get the FinetuneJob class for the provider."""
    provider_to_job_class = {
        _PROVIDER_ANYSCALE: FinetuneJobAnyScale,
        _PROVIDER_OPENAI: FinetuneJobOpenAI,
    }
    return provider_to_job_class[provider]


def get_provider_finetune_function(provider: str) -> callable:
    """Return the finetune function for the given model."""
    provider_to_finetune_function = {
        _PROVIDER_ANYSCALE: finetune_anyscale,
        _PROVIDER_OPENAI: finetune_openai,
    }
    return provider_to_finetune_function[provider]


# Note: Type of LM should be LM. We aren't importing it here to avoid
# circular imports.
def execute_finetune_job(job: FinetuneJob, lm: Any, cache_finetune: bool = True):
    """Execute the finetune job in a blocking manner."""
    try:
        job_kwargs = job.get_kwargs()
        if cache_finetune:
            model = cached_finetune(job=job, **job_kwargs)
        else:
            model = finetune(job=job, **job_kwargs)
        lm = lm.copy(model=model)
        job.set_result(lm)
    except Exception as err:
        logger.error(err)
        job.set_result(err)


# TODO: Add DiskCache, ignore job
def cached_finetune(
    job,
    model: str,
    train_data: List[Dict[str, Any]],
    train_kwargs: Optional[Dict[str, Any]] = None,
    train_method: TrainingMethod = TrainingMethod.SFT,
    provider: str = "openai",
) -> Union[str, Exception]:
    return finetune(
        job=job,
        model=model,
        train_data=train_data,
        train_kwargs=train_kwargs,
        train_method=train_method,
        provider=provider,
    )


def finetune(
    job,
    model: str,
    train_data: List[Dict[str, Any]],
    train_kwargs: Optional[Dict[str, Any]] = None,
    train_method: TrainingMethod = TrainingMethod.SFT,
    provider: str = "openai",
) -> Union[str, Exception]:
    """Fine-tune a new model based on the given model."""
    # Get the fine-tuning provider
    try:
        # Get the finetune function
        provider_finetune_function = get_provider_finetune_function(provider)

        # Fine-tune a new model based on the given model
        model = provider_finetune_function(
            job=job,
            model=model,
            train_data=train_data,
            train_kwargs=train_kwargs,
            train_method=train_method,
        )
    except Exception as err:
        raise err

    return model



# from functools import wraps

# import dsp


# # applied right to left (innermost first, like function calls)
# def compose_decorators(*decorators):
#     def decorator(func):
#         for decorator in decorators[::-1]:
#             func = decorator(func)
#         return func
#     return decorator


# def shallow_copy_example_args(func):
#     @wraps(func)
#     def wrapper(*args, **kwargs):
#         args = [dsp.Example(arg) if isinstance(arg, dsp.Example) else arg for arg in args]
#         kwargs = {key: dsp.Example(value) if isinstance(value, dsp.Example) else value for key, value in kwargs.items()}
#         return func(*args, **kwargs)
#     return wrapper


# transformation = shallow_copy_example_args
# # transformation = compose_decorators(handle_compilation, shallow_copy_example_args)



# def compiled(func):
#     def wrapper(*args, **kwargs):
#         is_to_be_compiled = True #decorator_kwargs.get('compile', False)
#         compiled_lm = dsp.settings.compiled_lm

#         if is_to_be_compiled and compiled_lm:
#             assert len(args) == 1, len(args)
#             example = args[0]

#             with dsp.settings.context(lm=compiled_lm, show_guidelines=False):
#                 old_demos = list(example.demos)
#                 example = func(example.copy(demos=[]), **kwargs)
#                 return example.copy(demos=old_demos)
        
#         with dsp.settings.context(compiling=True):
#             return func(*args, **kwargs)

#     return wrapper



# import inspect
# import json
# import random
# import string

# import requests


# class FuncInspector:
#   def __init__(self):
#     self.calls = []


#   def inspect_inner(self, func, function_calls):
#     def wrapper(*args, **kwargs):
#       result = func(*args, **kwargs)
#       self.merge_result(result, function_calls)
#       return result
#     return wrapper


#   def inspect_func(self, func):
#     def wrapper(*args, **kwargs):
#       result = func(*args, **kwargs)
#       stack = inspect.stack()
#       function_calls = []
#       for i in range(len(stack)):
#         if stack[i][3] == "<module>":
#           break
#         if stack[i][3] != "wrapper":
#           function_calls.append(stack[i][3])
#       function_calls.reverse()
#       result = self.inspect_inner(result, function_calls)
#       return result
#     return wrapper
  
  
#   def parse(self, obj, delete_empty=False):
#     if isinstance(obj, list):
#       for elem in obj:
#         self.parse(elem, delete_empty)
#     if isinstance(obj, dict):
#       to_delete = []
#       for key in obj:
#         if delete_empty and not obj[key] or key == "completions":
#           to_delete.append(key)
#         else:
#           self.parse(obj[key], delete_empty)
#       for key in to_delete:
#         obj.pop(key)


#   def merge_result(self, result, function_calls):
#     prev_list = self.calls
#     prev_call = {} if not prev_list else prev_list[-1]
#     for call in function_calls[:-1]:
#       if call not in prev_call:
#         prev_call = {call: []}
#         prev_list.append(prev_call)
#       prev_list = prev_call[call]
#       prev_call = {} if not prev_list else prev_list[-1]

#     example_obj = result[0]
#     self.parse(example_obj)
#     prev_list.append({ function_calls[-1]: example_obj })


#   def view_data(self):
#     chars = string.digits + string.ascii_lowercase
#     id = ''.join(random.choices(chars, k=8))

#     post_url = 'http://127.0.0.1:5000/log-item'
#     parsed_calls = self.calls.copy()
#     self.parse(parsed_calls, delete_empty=True)
#     data = {'id': id, 'content': parsed_calls}
#     response = requests.post(post_url, json=data)
    
#     if response.status_code == 201:
#       print('Data created successfully')
#     else:
#       print(f'Error sending data to server: {response.status_code}')
#       return

#     frontend_url = f"http://localhost:3000?id={id}"
#     print(f"View the data here, {frontend_url}")


#   def output_json(self, out_path):
#     f = open(out_path, "w")
#     json_object = json.dumps(self.calls, indent=2)
#     f.write(json_object)



from typing import Any, Dict, List, Optional
import json
import yaml
import os

from dspy.utils.logging import logger
from dspy.clients.finetune import (
    FinetuneJob,
    TrainingMethod,
    save_data,
)
from dspy.clients.openai import openai_data_validation

try:
    # AnyScale fine-tuning requires the following additional imports
    import anyscale
    from anyscale.job import JobConfig
except ImportError:
    anyscale = None


# List of training methods supported by AnyScale
TRAINING_METHODS_ANYSCALE = [
    TrainingMethod.SFT,
]

PROVIDER_ANYSCALE = "anyscale"


def is_anyscale_model(model: str) -> bool:
    """Check if the model is an AnyScale model."""
    # TODO: This needs to be implemented to support fine-tuning
    logger.info("Is AnyScale model is not implemented, returning False as a default to not break lm.py")
    return False


class FinetuneJobAnyScale(FinetuneJob):

    def __init__(self, *args, **kwargs):
        self.job_id = None
        self.model_names = None
        super().__init__(*args, **kwargs)


def finetune_anyscale(
        job: FinetuneJobAnyScale,
        model: str,
        train_data: List[Dict[str, Any]],
        train_kwargs: Optional[Dict[str, Any]]=None,
        train_method: TrainingMethod = TrainingMethod.SFT,
    ) -> str:
    """Start the finetune job."""
    train_kwargs = train_kwargs or {}
    assert "model" not in train_kwargs, "Model should not be in the train_kwargs"
    assert anyscale.__version__ >= "0.24.65", "Anyscale version >= 0.24.65 is required to use the dataset upload feature"
    assert all([x in train_kwargs for x in ["job_config_path", "llmforge_config_path"]]), "Both job_config_path and llmforge_config_path are required"
    train_kwargs_copy = train_kwargs.copy()
    train_kwargs_copy["model"] = model


    job_config_path = train_kwargs.get("job_config_path", None)
    llmforge_config_path = train_kwargs.get("llmforge_config_path", None)
    serve_config_path = train_kwargs.get("serve_config_path", None)

    if train_method not in TRAINING_METHODS_ANYSCALE:
        raise NotImplementedError(f"AnyScale can only support {TRAINING_METHODS_ANYSCALE} for the time being")

    if not verify_dataset(train_data):
        # TODO: Does AnyScale support text completion models?
        err = "[Finetune] Error: Unable to verify that the dataset is in the correct format."
        logger.error(err)
        raise RuntimeError(err)

    train_data_path = save_data(train_data, provider_name=PROVIDER_ANYSCALE)

    # TODO(Isaac): Figure out a better pattern
    job_config_temp = yaml.safe_load(open(job_config_path, "r"))
    remote_train_path = submit_data(train_path=train_data_path, job_config=job_config_temp)

    job_config = generate_config_files(train_path=remote_train_path, llmforge_config_path=llmforge_config_path, job_config_path=job_config_path, model=model)

    # Remove potential duplicate compute config

    job.job_id = start_remote_training(job_config=job_config)

    wait_for_training(job.job_id)

    model_info = get_model_info(job.job_id)

    # model_info[storage_uri] is a path to your cloud where the best(last if no eval) checkpoint weights are forwarded
    storage_uri = model_info["storage_uri"]

    lora_dynamic_path = storage_uri.split(model)[0]
    final_model_name = model + storage_uri.split(model)[1]
    
    if serve_config_path:
        update_serve_model_config(lora_dynamic_path, serve_config_path)
    job.model_names = [final_model_name]

    return "openai/" + final_model_name

def wait_for_training(job_id):
    """Wait for the training to complete."""
    logger.info("[Finetune] Waiting for training to complete...")
    anyscale.job.wait(id=job_id)
    logger.info("[Finetune] Training completed.")


def update_serve_model_config(lora_dynamic_path: str, serve_config_path: str):
    """Update the model config storage location with the job_id."""
    with open(serve_config_path, "r") as f:
        serve_config = yaml.safe_load(f)
    
    model_config_location = serve_config["applications"][0]["args"]["llm_configs"][0]
    
    with open(model_config_location, "r") as f:
        model_config = yaml.safe_load(f)

    model_config["lora_config"]["dynamic_lora_loading_path"] = lora_dynamic_path
    
    with open(model_config_location, "w") as f:
        yaml.safe_dump(model_config, f)
    

def verify_dataset(dataset: List[dict[str, Any]]) -> bool:
    """Verify the training arguments before starting training."""
    logger.info("[Finetune] Verifying dataset...")
    dataset_validation = openai_data_validation(dataset)

    if dataset_validation:
        logger.error(f"Dataset validation failed: {dataset_validation}")
        return False

    return True


def submit_data(train_path: str, job_config: Dict[str, Any]):
    """Upload the data to cloud storage."""
    logger.info("[Finetune] Submitting data to remote storage...")
    dataset_suffix = os.path.basename(train_path).split(".")[0]
    dataset_name = f"dataset-{job_config.get('name', dataset_suffix)}"
    train_path_remote = anyscale.llm.dataset.upload(train_path, name=dataset_name, cloud=job_config.get("cloud", None)).storage_uri
    logger.info(f"[Finetune] Data submitted. Remote train path: {train_path_remote}")

    return train_path_remote


def generate_config_files(train_path: str, llmforge_config_path: str, job_config_path: str, model: str):
    assert llmforge_config_path, "LLMForge config is required to generate the config files"
    assert job_config_path, "Job config is required to start the finetuning job"
    
    llmforge_config = yaml.safe_load(open(llmforge_config_path, "r"))
    job_config_dict = yaml.safe_load(open(job_config_path, "r"))
    
    llmforge_config["model_id"] = model
    llmforge_config["train_path"] = train_path
    llmforge_config = {k: v for k, v in llmforge_config.items() if v is not None}

    logger.info(f"Model config data: {llmforge_config}")
    yaml.safe_dump(llmforge_config, open(llmforge_config_path, "w"))

    if not job_config_dict.get("env_vars", None):
        job_config_dict["env_vars"] = {}

    for env_var in ["HF_TOKEN", "HF_HOME", "WANDB_API_KEY"]:
        if env_var not in job_config_dict["env_vars"] and os.environ.get(env_var, None):
            job_config_dict["env_vars"][env_var] = os.environ[env_var]
    

    job_config = JobConfig(**job_config_dict)


    return job_config


def start_remote_training(job_config) -> str:
    logger.info("[Finetune] Starting remote training...")
    job_id: str = anyscale.job.submit(job_config)
    logger.info(f"[Finetune] Remote training started. Job ID: {job_id}")
    return job_id


def wait_for_training(job_id):
    logger.info("Waiting for training to complete")
    anyscale.job.wait(id=job_id, timeout_s=18000)


def get_model_info(job_id):
    logger.info("[Finetune] Retrieving model information from Anyscale Models SDK...")
    info = anyscale.llm.model.get(job_id=job_id).to_dict()
    logger.info(f"[Finetune] Model info retrieved: {info}")
    return info

def read_jsonl(filename):
    with open(filename, "r") as f:
        return [json.loads(line) for line in f]



# TODO: This should move internally. Same for passage_match. dspy.metrics.answer_exact_match, dspy.metrics.answer_passage_match

import dsp


def answer_exact_match(example, pred, trace=None, frac=1.0):
    assert(type(example.answer) is str or type(example.answer) is list)
    
    if type(example.answer) is str:
        return dsp.answer_match(pred.answer, [example.answer], frac=frac)
    else: # type(example.answer) is list
        return dsp.answer_match(pred.answer, example.answer, frac=frac)

answer_exact_match_str = dsp.answer_match

def answer_passage_match(example, pred, trace=None):
    assert(type(example.answer) is str or type(example.answer) is list)
    
    if type(example.answer) is str:
        return dsp.passage_match(pred.context, [example.answer])
    else: # type(example.answer) is list
        return dsp.passage_match(pred.context, example.answer)



import functools
import os
from typing import List, Optional

import openai

import dspy
from dsp.modules.cache_utils import NotebookCacheMemory, cache_turn_on
from dsp.utils import dotdict

# Check for necessary libraries and suggest installation if not found.
try:
    import clickhouse_connect
except ImportError:
    raise ImportError(
        "The 'myscale' extra is required to use MyScaleRM. Install it with `pip install dspy-ai[myscale]`",
    )

# Verify the compatibility of the OpenAI library version installed.
try:
    major, minor, _ = map(int, openai.__version__.split("."))
    OPENAI_VERSION_COMPATIBLE = major >= 1 and minor >= 16
except Exception:
    OPENAI_VERSION_COMPATIBLE = False

if not OPENAI_VERSION_COMPATIBLE:
    raise ImportError(
        "An incompatible OpenAI library version is installed. Ensure you have version 1.16.1 or later.",
    )

# Attempt to handle specific OpenAI errors; fallback to general ones if necessary.
try:
    import openai.error

    ERRORS = (openai.error.RateLimitError, openai.error.ServiceUnavailableError, openai.error.APIError)
except Exception:
    ERRORS = (openai.RateLimitError, openai.APIError)


class MyScaleRM(dspy.Retrieve):
    """
    A retrieval module that uses MyScaleDB to return the top passages for a given query.

    MyScaleDB is a fork of ClickHouse that focuses on vector similarity search and full
    text search.  MyScaleRM is designed to facilitate easy retrieval of information from
    MyScaleDB using embeddings.  It supports embedding generation through either a local
    model or the OpenAI API. This class abstracts away the complexities of connecting to
    MyScaleDB, managing API keys, and processing queries to return semantically
    relevant results.

    Assumes that a table named `database.table` exists in MyScaleDB, and that the
    table has column named `vector_column` that stores vector data and a vector index has
    been created on this column. Other metadata are stored in `metadata_columns`.

    Args:
        client (clickhouse_connect.driver.client.Client): A client connection to the MyScaleDB.
        table (str): Name of the table within the database to perform queries against.
        database (str, optional): Name of the database to query within MyScaleDB.
        metadata_columns(List[str], optional): A list of columns to include in the results.
        vector_column (str, optional): The name of the column in the table that stores vector data.
        k (int, optional): The number of closest matches to retrieve for a given query.
        openai_api_key (str, optional): The API key for accessing OpenAI's services.
        model (str, optional): Specifies the particular OpenAI model to use for embedding generation.
        use_local_model (bool): Flag indicating whether a local model is used for embeddings.

    """

    def __init__(
        self,
        client: clickhouse_connect.driver.client.Client,
        table: str,
        database: str = "default",
        metadata_columns: List[str] = ["text"],
        vector_column: str = "vector",
        k: int = 3,
        openai_api_key: Optional[str] = None,
        openai_model: Optional[str] = None,
        local_embed_model: Optional[str] = None,
    ):
        self.client = client
        self.database = database
        self.table = table
        if not metadata_columns:
            raise ValueError("metadata_columns is required")
        self.metadata_columns = metadata_columns
        self.vector_column = vector_column
        self.k = k
        self.openai_api_key = openai_api_key
        self.model = openai_model
        self.use_local_model = False

        if local_embed_model:
            self.setup_local_model(local_embed_model)
        elif openai_api_key:
            os.environ["OPENAI_API_KEY"] = self.openai_api_key

    def setup_local_model(self, model_name: str):
        """
        Configures a local model for embedding generation, including model and tokenizer loading.

        Args:
            model_name: The name or path to the pre-trained model to load.

        Raises:
            ModuleNotFoundError: If necessary libraries (torch or transformers) are not installed.
        """
        try:
            import torch
            from transformers import AutoModel, AutoTokenizer
        except ImportError as exc:
            raise ModuleNotFoundError(
                """You need to install PyTorch and Hugging Face's transformers library to use a local embedding model.
                Install the pytorch using `pip install torch` and transformers using `pip install transformers` """,
            ) from exc

        try:
            self._local_embed_model = AutoModel.from_pretrained(model_name)
            self._local_tokenizer = AutoTokenizer.from_pretrained(model_name)
            self.use_local_model = True
        except Exception as e:
            raise ValueError(f"Failed to load model or tokenizer. Error: {str(e)}")

        if torch.cuda.is_available():
            self.device = torch.device("cuda:0")
        elif torch.backends.mps.is_available():
            self.device = torch.device("mps")
        else:
            self.device = torch.device("cpu")

        self._local_embed_model.to(self.device)

    @functools.lru_cache(maxsize=None if cache_turn_on else 0)
    @NotebookCacheMemory.cache
    def get_embeddings(self, query: str) -> List[float]:
        """
        Determines the appropriate source (OpenAI or local model) for embedding generation based on class configuration,
        and retrieves embeddings for the provided queries.

        Args:
            query: A query to generate embeddings for.

        Returns:
            A list of embeddings corresponding to the query in the input list.

        Raises:
            ValueError: If neither an OpenAI API key nor a local model has been configured.
        """
        if self.openai_api_key and self.model:
            return self._get_embeddings_from_openai(query)
        elif self.use_local_model:
            return self._get_embedding_from_local_model(query)
        else:
            raise ValueError("No valid method for obtaining embeddings is configured.")

    # TO DO Add this method as Util method outside MyScaleRM
    def _get_embeddings_from_openai(self, query: str) -> List[float]:
        """
        Uses the OpenAI API to generate embeddings for the given query.

        Args:
            query: A string for which to generate embeddings.

        Returns:
            A list containing the embedding of a query.
        """
        response = openai.embeddings.create(model=self.model, input=query)
        return response.data[0].embedding

    # TO DO Add this method as Util method outside MyScaleRM
    def _get_embedding_from_local_model(self, query: str) -> List[float]:
        """
        Generates embeddings for a single query using the configured local model.

        Args:
            query: The text query to generate an embedding for.

        Returns:
            A list representing the query's embedding.
        """
        import torch

        self._local_embed_model.eval()  # Ensure the model is in evaluation mode

        inputs = self._local_tokenizer(query, return_tensors="pt", padding=True, truncation=True).to(self.device)
        with torch.no_grad():
            output = self._local_embed_model(**inputs)

        return output.last_hidden_state.mean(dim=1).squeeze().numpy().tolist()

    def forward(self, user_query: str, k: Optional[int] = None) -> List[dotdict]:
        """
        Executes a retrieval operation based on a user's query and returns the top k relevant results.

        Args:
            user_query: The query text to search for.
            k: Optional; The number of top matches to return. Defaults to the class's configured k value.

        Returns:
            A list of dotdict objects containing the formatted retrieval results.

        Raises:
            ValueError: If the user_query is None.
        """
        if user_query is None:
            raise ValueError("Query is required")
        k = k or self.k
        embeddings = self.get_embeddings(user_query)
        columns_string = ", ".join(self.metadata_columns)
        result = self.client.query(f"""
        SELECT {columns_string},
        distance({self.vector_column}, {embeddings}) as dist FROM {self.database}.{self.table} ORDER BY dist LIMIT {k}
        """)
        result = self.client.query(f"""
            SELECT {columns_string}, distance({self.vector_column}, {embeddings}) AS dist
            FROM {self.database}.{self.table}
            ORDER BY dist
            LIMIT {k}
        """)

        # Convert the metadata into strings to pass to dspy.Prediction
        results = []
        for row in result.named_results():
            if len(self.metadata_columns) == 1:
                results.append(row[self.metadata_columns[0]])
            else:
                row_strings = [f"{column}: {row[column]}" for column in self.metadata_columns]  # Format row data
                row_string = "\n".join(row_strings)  # Combine formatted data
                results.append(row_string)  # Append to results

        # Return list of dotdict
        return [dotdict({"long_text": passage}) for passage in results]



from typing import List, Optional

import numpy as np

import dsp


class KNN:
    def __init__(self, k: int, trainset: List[dsp.Example], vectorizer: Optional[dsp.BaseSentenceVectorizer] = None):
        self.k = k
        self.trainset = trainset
        self.vectorizer = vectorizer or dsp.SentenceTransformersVectorizer()
        trainset_casted_to_vectorize = [
            " | ".join([f"{key}: {value}" for key, value in example.items() if key in example._input_keys])
            for example in self.trainset
        ]
        self.trainset_vectors = self.vectorizer(trainset_casted_to_vectorize).astype(np.float32)

    def __call__(self, **kwargs) -> List[dsp.Example]:
        with dsp.settings.context(vectorizer=self.vectorizer):
            input_example_vector = self.vectorizer([" | ".join([f"{key}: {val}" for key, val in kwargs.items()])])
            scores = np.dot(self.trainset_vectors, input_example_vector.T).squeeze()
            nearest_samples_idxs = scores.argsort()[-self.k :][::-1]
            train_sampled = [self.trainset[cur_idx] for cur_idx in nearest_samples_idxs]
            return train_sampled



"""
Retriever model for deeplake
"""

from collections import defaultdict
from typing import List, Optional, Union

import openai

import dspy
from dsp.utils import dotdict

try:
    import openai

    ERRORS = (
        openai.RateLimitError,
        openai.APIError,
    )
except Exception:
    ERRORS = (openai.RateLimitError, openai.APIError)


class DeeplakeRM(dspy.Retrieve):
    
    """
    A retriever module that uses deeplake to return the top passages for a given query.

    Assumes that a Deep Lake Vector Store has been created and populated with the following payload:
        - text: The text of the passage

    Args:
        deeplake_vectorstore_name (str): The name or path of the Deep Lake Vector Store.
        deeplake_client (VectorStore): An instance of the Deep Lake client.
        k (int, optional): The default number of top passages to retrieve. Defaults to 3.

    Examples:
        Below is a code snippet that shows how to use Deep Lake as the default retriver:
        ```python
        from deeplake import VectorStore
        llm = dspy.OpenAI(model="gpt-3.5-turbo")
        deeplake_client = VectorStore
        retriever_model = DeeplakeRM("my_vectorstore_path", deeplake_client=deeplake_client)
        dspy.settings.configure(lm=llm, rm=retriever_model)
        ```

        Below is a code snippet that shows how to use Deep Lake in the forward() function of a module
        ```python
        self.retrieve = DeeplakeRM("my_vectorstore_path", deeplake_client=deeplake_client, k=num_passages)
        ```
    """

    def __init__(
        self,
        deeplake_vectorstore_name: str,
        deeplake_client,
        k: int = 3,
    ):
        try:
            from deeplake import VectorStore
        except ImportError:
            raise ImportError("The 'deeplake' extra is required to use DeepLakeRM. Install it with `pip install dspy-ai[deeplake]`",)

        self._deeplake_vectorstore_name = deeplake_vectorstore_name
        self._deeplake_client = deeplake_client(
            path=self._deeplake_vectorstore_name,
            embedding_function=self.embedding_function,
            )

        super().__init__(k=k)

    def embedding_function(self, texts, model="text-embedding-ada-002"):
        if isinstance(texts, str):
            texts = [texts]

        texts = [t.replace("\n", " ") for t in texts]
        
        return [data.embedding for data in openai.embeddings.create(input = texts, model=model).data]
    
    def forward(
        self, query_or_queries: Union[str, List[str]], k: Optional[int],**kwargs,
    ) -> dspy.Prediction:
        
        """Search with DeepLake for self.k top passages for query

        Args:
            query_or_queries (Union[str, List[str]]): The query or queries to search for.
            k (Optional[int]): The number of top passages to retrieve. Defaults to self.k.

        Returns:
            dspy.Prediction: An object containing the retrieved passages.
        """
        queries = (
            [query_or_queries]
            if isinstance(query_or_queries, str)
            else query_or_queries
        )
        queries = [q for q in queries if q]  # Filter empty queries

        k = k if k is not None else self.k

        passages = defaultdict(float)
        #deeplake doesn't support batch querying, manually querying each query and storing them
        for query in queries:
            results = self._deeplake_client.search(query, k=k, **kwargs)

            for score,text in zip(results.get('score',0.0),results.get('text',"")):
                passages[text] += score

        sorted_passages = sorted(
            passages.items(), key=lambda x: x[1], reverse=True)[:k]
        
        return [dotdict({"long_text": p}) for p, _ in sorted_passages]


from dsp.utils import normalize_text
from dspy.primitives.prediction import Completions, Prediction

default_normalize = lambda s: normalize_text(s) or None


def majority(prediction_or_completions, normalize=default_normalize, field=None):
    """
        Returns the most common completion for the target field (or the last field) in the signature.
        When normalize returns None, that completion is ignored.
        In case of a tie, earlier completion are prioritized.
    """

    assert any(isinstance(prediction_or_completions, t) for t in [Prediction, Completions, list])
    input_type = type(prediction_or_completions)

    # Get the completions
    if isinstance(prediction_or_completions, Prediction):
        completions = prediction_or_completions.completions
    else:
        completions = prediction_or_completions
    
    try:
        signature = completions.signature
    except:
        signature = None
    
    if not field:
        if signature:
            field = signature.output_fields[-1]
        else:
            field = list(completions[0].keys())[-1]

    # Normalize
    normalize = normalize if normalize else lambda x: x
    normalized_values = [normalize(completion[field]) for completion in completions]
    normalized_values_ = [x for x in normalized_values if x is not None]
    
    # Count
    value_counts = {}
    for value in (normalized_values_ or normalized_values):
        value_counts[value] = value_counts.get(value, 0) + 1

    majority_value = max(value_counts, key=value_counts.get)

    # Return the first completion with the majority value in the field
    for completion in completions:
        if normalize(completion[field]) == majority_value:
            break
    
    # if input_type == Prediction:
    return Prediction.from_completions([completion], signature=signature)





from typing import Any

from dsp.modules.lm import LM

ibm_watsonx_ai_api_error = False

try:
    import ibm_watsonx_ai  # noqa: F401
    from ibm_watsonx_ai.foundation_models import Model  # type: ignore

except ImportError:
    ibm_watsonx_ai_api_error = Exception


class Watsonx(LM):
    """Wrapper around Watsonx AI's API.

    The constructor initializes the base class LM to support prompting requests to Watsonx models.
    This requires the following parameters:
    Args:
        model (str): the type of model to use from IBM Watsonx AI.
        credentials ([dict]): credentials to Watson Machine Learning instance.
        project_id (str): ID of the Watson Studio project.
        **kwargs: Additional arguments to pass to the API provider. This is initialized with default values for relevant
            text generation parameters needed for communicating with Watsonx API, such as:
                - decoding_method
                - max_new_tokens
                - min_new_tokens
                - stop_sequences
                - repetition_penalty
    """

    def __init__(self, model, credentials, project_id, **kwargs):
        """Parameters

        model : str
            Which pre-trained model from Watsonx.ai to use?
            Choices are [
                `mistralai/mixtral-8x7b-instruct-v01`,
                `ibm/granite-13b-instruct-v2`,
                `meta-llama/llama-3-70b-instruct`]
        credentials : [dict]
            Credentials to Watson Machine Learning instance.
        project_id : str
            ID of the Watson Studio project.
        **kwargs: dict
            Additional arguments to pass to the API provider.
        """
        self.model = model
        self.credentials = credentials
        self.project_id = project_id
        self.provider = "ibm"
        self.model_type = "instruct"
        self.kwargs = {
            "temperature": 0,
            "decoding_method": "greedy",
            "max_new_tokens": 150,
            "min_new_tokens": 0,
            "stop_sequences": [],
            "repetition_penalty": 1,
            "num_generations": 1,
            **kwargs,
        }

        self.client = Model(
            model_id=self.model,
            params=self.kwargs,
            credentials=self.credentials,
            project_id=self.project_id,
        )

        self.history: list[dict[str, Any]] = []

    def basic_request(self, prompt: str, **kwargs) -> Any:
        raw_kwargs = kwargs
        kwargs = {**self.kwargs, **kwargs}

        response = self.client.generate(prompt, params={**kwargs})

        history = {
            "prompt": prompt,
            "response": response,
            "kwargs": kwargs,
            "raw_kwargs": raw_kwargs,
        }
        self.history.append(history)

        return response

    def request(self, prompt: str, **kwargs) -> Any:
        # Handles the specific prompting for each supported model and the retrieval of completions from IBM Watsonx AI

        if self.model == "mistralai/mixtral-8x7b-instruct-v01":
            prompt = "<s>[INST]" + prompt + "</INST>"
        elif self.model == "meta-llama/llama-3-70b-instruct":
            prompt = (
                "<|begin_of_text|><|start_header_id|>system<|end_header_id|>"
                + prompt
                + "<|eot_id|><|start_header_id|>user<|end_header_id|>"
            )

        return self.basic_request(prompt, **kwargs)

    def __call__(
        self,
        prompt: str,
        only_completed: bool = True,
        return_sorted: bool = False,
        **kwargs,
    ) -> list[dict[str, Any]]:
        """Retrieves completions from Watsonx.

        Args:
            prompt (str): prompt to send to Watsonx
            only_completed (bool, optional): return only completed responses and ignores completion due to length.
            Defaults to True.
            return_sorted (bool, optional): sort the completion choices using the returned probabilities.
            Defaults to False.
            **kwargs: Additional arguments to pass

        Returns:
            list[dict[str, Any]]: list of completion choices
        """
        if only_completed is False:
            raise ValueError("only_completed is True for now")

        if return_sorted:
            raise ValueError("return_sorted is False for now")

        response = self.request(prompt, **kwargs)

        return [result["generated_text"] for result in response["results"]]



import logging
import os
import sys
import typing as t

import structlog

logger = structlog.get_logger()


class LogSettings:
    def __init__(self, output_type: str, method: str, file_name: t.Optional[str]) -> None:
        self.output_type = output_type
        self.method = method
        self.file_name = file_name
        self._configure_structlog()

    def _configure_structlog(self):
        if self.output_type == "str":
            renderer = structlog.dev.ConsoleRenderer()
        else:
            renderer = structlog.processors.JSONRenderer()

        structlog.configure(
            processors=[
                structlog.stdlib.add_logger_name,
                structlog.stdlib.add_log_level,
                structlog.processors.CallsiteParameterAdder(
                    {
                        structlog.processors.CallsiteParameter.FILENAME,
                        structlog.processors.CallsiteParameter.LINENO,
                    },
                ),
                structlog.processors.TimeStamper(fmt="iso"),
                structlog.processors.StackInfoRenderer(),
                structlog.processors.format_exc_info,
                structlog.processors.UnicodeDecoder(),
                renderer,
            ],
            logger_factory=structlog.stdlib.LoggerFactory(),
            wrapper_class=structlog.stdlib.BoundLogger,
        )

    def set_log_output(
        self,
        method: t.Optional[str] = None,
        file_name: t.Optional[str] = None,
        output_type: t.Optional[str] = None,
    ):
        if method is not None and method not in ["console", "file"]:
            raise ValueError("method provided can only be 'console', 'file'")

        if method == "file" and file_name is None:
            raise ValueError("file_name must be provided when method = 'file'")

        if method is not None:
            self.method = method
            self.file_name = file_name

        if output_type is not None and output_type not in ["str", "json"]:
            raise ValueError("output_type provided can only be 'str', 'json'")

        if output_type is not None:
            self.output_type = output_type

        # Update Renderer
        self._configure_structlog()

        # Grab the root logger
        log = logging.getLogger()
        for handler in log.handlers[:]:
            log.removeHandler(handler)

        # Add new Handler
        if self.method == "file":
            assert self.file_name is not None
            log.addHandler(logging.FileHandler(self.file_name))
        else:
            log.addHandler(logging.StreamHandler(sys.stdout))


level = os.environ.get("LOG_LEVEL", "info").upper()


# Set Defaults
def show_logging(level: str = level) -> None:
    logging.basicConfig(
        format="%(message)s",
        stream=sys.stdout,
        level=level,
    )

    # 'backoff' is used by OpenAI, and defaults their log level to INFO.
    # this can clobber up dspy relevant logging for users
    # this silences there logs.
    logging.getLogger("backoff").setLevel(logging.WARNING)


settings = LogSettings(output_type="str", method="console", file_name=None)
set_log_output = settings.set_log_output



import dspy
import inspect

from pydantic import BaseModel
from dspy.primitives.program import Module
from dspy.signatures.signature import ensure_signature
from dspy.adapters.json_adapter import get_annotation_name
from typing import Callable, Any, get_type_hints, get_origin, Literal

class Tool:
    def __init__(self, func: Callable, name: str = None, desc: str = None, args: dict[str, Any] = None):
        annotations_func = func if inspect.isfunction(func) else func.__call__
        self.func = func
        self.name = name or getattr(func, '__name__', type(func).__name__)
        self.desc = desc or getattr(func, '__doc__', None) or getattr(annotations_func, '__doc__', "No description")
        self.args = {
            k: v.schema() if isinstance((origin := get_origin(v) or v), type) and issubclass(origin, BaseModel)
            else get_annotation_name(v)
            for k, v in (args or get_type_hints(annotations_func)).items() if k != 'return'
        }

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)


class ReAct(Module):
    def __init__(self, signature, tools: list[Callable], max_iters=5):
        """
        Tools is either a list of functions, callable classes, or dspy.Tool instances.
        """

        self.signature = signature = ensure_signature(signature)
        self.max_iters = max_iters

        tools = [t if isinstance(t, Tool) or hasattr(t, 'input_variable') else Tool(t) for t in tools]
        tools = {tool.name: tool for tool in tools}

        inputs_ = ", ".join([f"`{k}`" for k in signature.input_fields.keys()])
        outputs_ = ", ".join([f"`{k}`" for k in signature.output_fields.keys()])
        instr = [f"{signature.instructions}\n"] if signature.instructions else []

        instr.extend([
            f"You will be given {inputs_} and your goal is to finish with {outputs_}.\n",
            "To do this, you will interleave Thought, Tool Name, and Tool Args, and receive a resulting Observation.\n",
            "Thought can reason about the current situation, and Tool Name can be the following types:\n",
        ])

        finish_desc = f"Signals that the final outputs, i.e. {outputs_}, are now available and marks the task as complete."
        finish_args = {} #k: v.annotation for k, v in signature.output_fields.items()}
        tools["finish"] = Tool(func=lambda **kwargs: "Completed.", name="finish", desc=finish_desc, args=finish_args)

        for idx, tool in enumerate(tools.values()):
            desc = tool.desc.replace("\n", "  ")
            args = tool.args if hasattr(tool, 'args') else str({tool.input_variable: str})
            desc = f"whose description is <desc>{desc}</desc>. It takes arguments {args} in JSON format."
            instr.append(f"({idx+1}) {tool.name}, {desc}")

        signature_ = (
            dspy.Signature({**signature.input_fields}, "\n".join(instr))
            .append("trajectory", dspy.InputField(), type_=str)
            .append("next_thought", dspy.OutputField(), type_=str)
            .append("next_tool_name", dspy.OutputField(), type_=Literal[tuple(tools.keys())])
            .append("next_tool_args", dspy.OutputField(), type_=dict[str, Any])
        )

        fallback_signature = (
            dspy.Signature({**signature.input_fields, **signature.output_fields})
            .append("trajectory", dspy.InputField(), type_=str)
        )

        self.tools = tools
        self.react = dspy.Predict(signature_)
        self.extract = dspy.ChainOfThought(fallback_signature)
    
    def forward(self, **input_args):
        trajectory = {}

        def format(trajectory_: dict[str, Any], last_iteration: bool):
            adapter = dspy.settings.adapter or dspy.ChatAdapter()
            signature_ = dspy.Signature(f"{', '.join(trajectory_.keys())} -> x")
            return adapter.format_fields(signature_, trajectory_, role='user')

        for idx in range(self.max_iters):
            pred = self.react(**input_args, trajectory=format(trajectory, last_iteration=(idx == self.max_iters-1)))

            trajectory[f"thought_{idx}"] = pred.next_thought
            trajectory[f"tool_name_{idx}"] = pred.next_tool_name
            trajectory[f"tool_args_{idx}"] = pred.next_tool_args

            try:
                trajectory[f"observation_{idx}"] = self.tools[pred.next_tool_name](**pred.next_tool_args)
            except Exception as e:
                trajectory[f"observation_{idx}"] = f"Failed to execute: {e}"

            if pred.next_tool_name == "finish":
                break
        
        extract = self.extract(**input_args, trajectory=format(trajectory, last_iteration=False))
        return dspy.Prediction(trajectory=trajectory, **extract)



import os
from abc import abstractmethod
from concurrent.futures import Future
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional
from dspy.utils.logging import logger

import ujson
from datasets.fingerprint import Hasher


def get_finetune_directory() -> str:
    """Get the directory to save the fine-tuned models."""
    # TODO: Move to a centralized location with all the other env variables
    dspy_cachedir = os.environ.get("DSPY_CACHEDIR")
    dspy_cachedir = dspy_cachedir or os.path.join(Path.home(), ".dspy_cache")
    finetune_dir = os.path.join(dspy_cachedir, "finetune")
    finetune_dir = os.path.abspath(finetune_dir)
    return finetune_dir


FINETUNE_DIRECTORY = get_finetune_directory()


class TrainingMethod(str, Enum):
    """Enum class for training methods.

    When comparing enums, Python checks for object IDs, which means that the
    enums can't be compared directly. Subclassing the Enum class along with the
    str class allows for direct comparison of the enums.
    """

    SFT = "SFT"
    Preference = "Preference"


class TrainingStatus(str, Enum):
    """Enum class for remote training status."""

    not_started = "not_started"
    pending = "pending"
    running = "running"
    succeeded = "succeeded"
    failed = "failed"
    cancelled = "cancelled"


"""Dictionary mapping training methods to the data keys they require."""
TRAINING_METHOD_TO_DATA_KEYS = {
    TrainingMethod.SFT: ["prompt", "completion"],
    TrainingMethod.Preference: ["prompt", "chosen", "rejected"],
}


class FinetuneJob(Future):
    def __init__(
        self,
        model: str,
        train_data: List[Dict[str, Any]],
        train_kwargs: Optional[Dict[str, Any]] = None,
        train_method: TrainingMethod = TrainingMethod.SFT,
        provider: str = "openai",
    ):
        self.model = model
        self.train_data = train_data
        self.train_kwargs: Dict[str, Any] = train_kwargs or {}
        self.train_method = train_method
        self.provider = provider
        super().__init__()

    def get_kwargs(self):
        return dict(
            model=self.model,
            train_data=self.train_data,
            train_kwargs=self.train_kwargs,
            train_method=self.train_method,
            provider=self.provider,
        )

    def __repr__(self):
        return str(self)

    # Subclasses should override the cancel method to cancel the finetune job;
    # then call the super's cancel method so that the future can be cancelled.
    def cancel(self):
        """Cancel the finetune job."""
        super().cancel()

    @abstractmethod
    def status(self):
        """Get the status of the finetune job."""
        raise NotImplementedError("Method `status` is not implemented.")


def validate_finetune_data(data: List[Dict[str, Any]], train_method: TrainingMethod):
    """Validate the finetune data based on the training method."""
    # Get the required data keys for the training method
    required_keys = TRAINING_METHOD_TO_DATA_KEYS[train_method]

    # Check if the training data has the required keys
    for ind, data_dict in enumerate(data):
        if not all([key in data_dict for key in required_keys]):
            raise ValueError(
                f"The datapoint at index {ind} is missing the keys required for {train_method} training. Expected: "
                f"{required_keys}, Found: {data_dict.keys()}"
            )


def save_data(
    data: List[Dict[str, Any]],
    provider_name: Optional[str] = None,
) -> str:
    """Save the fine-tuning data to a file."""
    logger.info("[Finetune] Converting data to JSONL format...")
    # Construct the file name based on the data hash
    hash = Hasher.hash(data)
    file_name = f"{hash}.jsonl"
    file_name = f"{provider_name}_{file_name}" if provider_name else file_name

    # Find the directory to save the fine-tuning data
    finetune_parent_dir = get_finetune_directory()
    os.makedirs(finetune_parent_dir, exist_ok=True)

    # Save the data to a file
    file_path = os.path.join(finetune_parent_dir, file_name)
    file_path = os.path.abspath(file_path)
    with open(file_path, "w") as f:
        for item in data:
            f.write(ujson.dumps(item) + "\n")
    return file_path



import random

from datasets import load_dataset

from dspy.datasets.dataset import Dataset


class HotPotQA(Dataset):
    def __init__(self, *args, only_hard_examples=True, keep_details='dev_titles', unofficial_dev=True, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        assert only_hard_examples, "Care must be taken when adding support for easy examples." \
                                   "Dev must be all hard to match official dev, but training can be flexible."
        
        hf_official_train = load_dataset("hotpot_qa", 'fullwiki', split='train', trust_remote_code=True)
        hf_official_dev = load_dataset("hotpot_qa", 'fullwiki', split='validation', trust_remote_code=True)

        official_train = []
        for raw_example in hf_official_train:
            if raw_example['level'] == 'hard':
                if keep_details is True:
                    keys = ['id', 'question', 'answer', 'type', 'supporting_facts', 'context']
                elif keep_details == 'dev_titles':
                    keys = ['question', 'answer', 'supporting_facts']
                else:
                    keys = ['question', 'answer']

                example = {k: raw_example[k] for k in keys}
                
                if 'supporting_facts' in example:
                    example['gold_titles'] = set(example['supporting_facts']['title'])
                    del example['supporting_facts']

                official_train.append(example)

        rng = random.Random(0)
        rng.shuffle(official_train)

        self._train = official_train[:len(official_train)*75//100]

        if unofficial_dev:
            self._dev = official_train[len(official_train)*75//100:]
        else:
            self._dev = None

        for example in self._train:
            if keep_details == 'dev_titles':
                del example['gold_titles']
        
        test = []
        for raw_example in hf_official_dev:
            assert raw_example['level'] == 'hard'
            example = {k: raw_example[k] for k in ['id', 'question', 'answer', 'type', 'supporting_facts']}
            if 'supporting_facts' in example:
                example['gold_titles'] = set(example['supporting_facts']['title'])
                del example['supporting_facts']
            test.append(example)

        self._test = test


if __name__ == '__main__':
    from dsp.utils import dotdict

    data_args = dotdict(train_seed=1, train_size=16, eval_seed=2023, dev_size=200*5, test_size=0)
    dataset = HotPotQA(**data_args)

    print(dataset)
    print(dataset.train[0].question)
    print(dataset.train[15].question)

    print(len(dataset.train), len(dataset.dev), len(dataset.test))

    print(dataset.dev[0].question)
    print(dataset.dev[340].question)
    print(dataset.dev[937].question)

"""
What was the population of the city where Woodward Avenue ends in 2010?
Where did the star , who is also an executive producer, of the Mick begin her carrer? 
16 1000 0
Both London and German have seen attacks during war, there was one specific type of attack that Germany called the blitz, what did London call a similar attack?
Pre-Madonna was a collection of demos by the singer who was a leading presence during the emergence of what network?
Alan Mills composed the classic folk song that tells the story of what? 
"""



from collections import defaultdict
from typing import Optional, Union

import dspy
from dsp.modules.sentence_vectorizer import BaseSentenceVectorizer, FastEmbedVectorizer
from dsp.utils import dotdict

try:
    from qdrant_client import QdrantClient, models
except ImportError as e:
    raise ImportError(
        "The 'qdrant' extra is required to use QdrantRM. Install it with `pip install dspy-ai[qdrant]`",
    ) from e


class QdrantRM(dspy.Retrieve):
    """A retrieval module that uses Qdrant to return the top passages for a given query.

    Args:
        qdrant_collection_name (str): The name of the Qdrant collection.
        qdrant_client (QdrantClient): An instance of `qdrant_client.QdrantClient`.
        k (int, optional): The default number of top passages to retrieve. Default: 3.
        document_field (str, optional): The key in the Qdrant payload with the content. Default: `"document"`.
        vectorizer (BaseSentenceVectorizer, optional): An implementation `BaseSentenceVectorizer`.
                                                           Default: `FastEmbedVectorizer`.
        vector_name (str, optional): Name of the vector in the collection. Default: The first available vector name.

    Examples:
        Below is a code snippet that shows how to use Qdrant as the default retriver:
        ```python
        from qdrant_client import QdrantClient

        llm = dspy.OpenAI(model="gpt-3.5-turbo")
        qdrant_client = QdrantClient()
        retriever_model = QdrantRM("my_collection_name", qdrant_client=qdrant_client)
        dspy.settings.configure(lm=llm, rm=retriever_model)
        ```

        Below is a code snippet that shows how to use Qdrant in the forward() function of a module
        ```python
        self.retrieve = QdrantRM(question, k=num_passages, filter=filter)
        ```
    """

    def __init__(
        self,
        qdrant_collection_name: str,
        qdrant_client: QdrantClient,
        k: int = 3,
        document_field: str = "document",
        vectorizer: Optional[BaseSentenceVectorizer] = None,
        vector_name: Optional[str] = None,
    ):
        self._collection_name = qdrant_collection_name
        self._client = qdrant_client

        self._vectorizer = vectorizer or FastEmbedVectorizer(self._client.embedding_model_name)

        self._document_field = document_field

        self._vector_name = vector_name or self._get_first_vector_name()

        super().__init__(k=k)

    def forward(self, query_or_queries: Union[str, list[str]], k: Optional[int] = None, filter: Optional[models.Filter]=None) -> dspy.Prediction:
        """Search with Qdrant for self.k top passages for query.

        Args:
            query_or_queries (Union[str, List[str]]): The query or queries to search for.
            k (Optional[int]): The number of top passages to retrieve. Defaults to self.k.
            filter (Optional["Filter"]): "Look only for points which satisfies this conditions". Default: None.

        Returns:
            dspy.Prediction: An object containing the retrieved passages.
        """
        queries = [query_or_queries] if isinstance(query_or_queries, str) else query_or_queries
        queries = [q for q in queries if q]  # Filter empty queries

        vectors = self._vectorizer(queries)

        search_requests = [
            models.QueryRequest(
                query=vector,
                using=self._vector_name,
                limit=k or self.k,
                with_payload=[self._document_field],
                filter=filter,
            )
            for vector in vectors
        ]
        batch_results = self._client.query_batch_points(self._collection_name, requests=search_requests)

        passages_scores = defaultdict(float)
        for batch in batch_results:
            for result in batch.points:
                # If a passage is returned multiple times, the score is accumulated.
                document = result.payload.get(self._document_field)
                passages_scores[document] += result.score

        # Sort passages by their accumulated scores in descending order
        sorted_passages = sorted(passages_scores.items(), key=lambda x: x[1], reverse=True)[:k]

        # Wrap each sorted passage in a dotdict with 'long_text'
        return [dotdict({"long_text": passage}) for passage, _ in sorted_passages]

    def _get_first_vector_name(self) -> Optional[str]:
        vectors = self._client.get_collection(self._collection_name).config.params.vectors

        if not isinstance(vectors, dict):
            # The collection only has the default, unnamed vector
            return None

        first_vector_name = list(vectors.keys())[0]

        # The collection has multiple vectors. Could also include the falsy unnamed vector - Empty string("")
        return first_vector_name or None



import os
from typing import Any, Literal, Optional

import requests

from dsp.modules.lm import LM

SMART_ENDPOINT = "https://chat-api.you.com/smart"
RESEARCH_ENDPOINT = "https://chat-api.you.com/research"


class You(LM):
    """Wrapper around You.com's conversational Smart and Research APIs.

    Each API endpoint is designed to generate conversational
    responses to a variety of query types, including inline citations
    and web results when relevant.

    Smart Mode:
    - Quick, reliable answers for a variety of questions
    - Cites the entire web page URL

    Research Mode:
    - In-depth answers with extensive citations for a variety of questions
    - Cites the specific web page snippet relevant to the claim

    To connect to the You.com api requires an API key which
    you can get at https://api.you.com.

    For more information, check out the documentations at
    https://documentation.you.com/api-reference/.

    Args:
        endpoint: You.com conversational endpoints. Choose from "smart" or "research"
        api_key: You.com API key, if `YDC_API_KEY` is not set in the environment
    """

    def __init__(
        self,
        endpoint: Literal["smart", "research"] = "smart",
        ydc_api_key: Optional[str] = None,
    ):
        super().__init__(model="you.com")
        self.ydc_api_key = ydc_api_key or os.environ["YDC_API_KEY"]
        self.endpoint = endpoint

        # Mandatory DSPy attributes to inspect LLM call history
        self.history = []
        self.provider = "you.com"

    def basic_request(self, prompt, **kwargs) -> dict[str, Any]:
        headers = {"x-api-key": self.ydc_api_key}
        params = {"query": prompt}  # DSPy `kwargs` are ignored as they are not supported by the API

        response = requests.post(self.request_endpoint, headers=headers, json=params)
        response.raise_for_status()

        data = response.json()

        # Update history
        self.history.append({"prompt": prompt, "response": data, "endpoint": self.endpoint})

        return data

    @property
    def request_endpoint(self) -> str:
        if self.endpoint == "smart":
            return SMART_ENDPOINT
        return RESEARCH_ENDPOINT

    def __call__(self, prompt, only_completed: bool = True, return_sorted: bool = False, **kwargs) -> list[str]:
        response = self.request(prompt, **kwargs)
        return [response["answer"]]



# Adapted from: https://www.philschmid.de/fine-tune-flan-t5#3-fine-tune-and-evaluate-flan-t5

import copy
import glob
import json
import os
import warnings
from dataclasses import dataclass

import evaluate
import numpy as np
import torch
from datasets import Dataset
from transformers import (
    AutoConfig,
    AutoModelForCausalLM,
    AutoModelForSeq2SeqLM,
    AutoTokenizer,
    DataCollatorForSeq2Seq,
    PreTrainedTokenizer,
    Seq2SeqTrainer,
    Seq2SeqTrainingArguments,
    Trainer,
    TrainingArguments,
    set_seed,
)

# from peft import get_peft_model, LoraConfig, TaskType
from transformers.trainer_callback import TrainerCallback

# from dsp.modules.finetuning.fid import *


warnings.filterwarnings("ignore")

IGNORE_INDEX = -100
DEFAULT_SEP_TOKEN = "[SEP]"
DEFAULT_PAD_TOKEN = "[PAD]"
DEFAULT_EOS_TOKEN = "</s>"
DEFAULT_BOS_TOKEN = "<s>"
DEFAULT_UNK_TOKEN = "</s>"
SPECIAL_TOKENS_DICT = {
    "sep_token": DEFAULT_SEP_TOKEN,
    "pad_token": DEFAULT_PAD_TOKEN,
    # "eos_token": DEFAULT_EOS_TOKEN,
    # "bos_token": DEFAULT_BOS_TOKEN,
    "unk_token": DEFAULT_UNK_TOKEN,
}


def _freeze_model_layers(model, unfreeze_last_n):
    # Freeze all layers
    for parameter in model.parameters():
        parameter.requires_grad = False

    # Unfreeze the last n transformer blocks in the decoder
    NUM_DECODER_LAYERS = len(model.transformer.h)
    for i, m in enumerate(model.transformer.h):
        if i >= NUM_DECODER_LAYERS - unfreeze_last_n:
            for parameter in m.parameters():
                parameter.requires_grad = True 

    # Unfreeze parameters after decoder block
    for parameter in model.transformer.ln_f.parameters():
        parameter.requires_grad = True
    for parameter in model.lm_head.parameters():        
        parameter.requires_grad = True
    return model


def _load_data(path):
    # dataset = Dataset.from_json(path)
    L = []
    import ujson
    with open(path) as f:
        for line in f:
            L.append(ujson.loads(line))

    dataset = Dataset.from_list(L)
    return dataset


def preprocess_prompt(text, tokenizer, encoder_decoder_model, decoder_only_model, rationale):
    text = f'{text} ' if encoder_decoder_model else f'{text} {tokenizer.sep_token}'
    return text


def preprocess_completion(text, tokenizer, encoder_decoder_model, decoder_only_model, rationale):
    text = text if encoder_decoder_model else f'{text}{tokenizer.eos_token}'
    return text.lstrip()


def _preprocess_data(dataset, tokenizer, encoder_decoder_model, decoder_only_model, config):
    dataset = dataset.map(lambda x: {
        "prompt": preprocess_prompt(x["prompt"], tokenizer, encoder_decoder_model, decoder_only_model, config['rationale']),
        "completion": preprocess_completion(x["completion"], tokenizer, encoder_decoder_model, decoder_only_model, config['rationale']),
    })
    skipped = [x for x in dataset if x["completion"] is None]
    print(f'# examples skipped due to parsing error: {len(skipped)} / {len(dataset)}')
    dataset = dataset.filter(lambda x: x["completion"])
    return dataset


def _tokenize_dataset(dataset, tokenizer, encoder_decoder_model, decoder_only_model):
    def get_dataset_stats(dataset, tokenizer, column):
        tokenized_inputs = dataset.map(lambda x: tokenizer(x[column]), batched=True)
        max_length = max([len(x) for x in tokenized_inputs["input_ids"]])
        return max_length

    def get_tokens_seq2seq(sample, max_source_length, max_target_length, padding="max_length"):
        # Tokenize inputs
        model_inputs = tokenizer(sample["prompt"], max_length=max_source_length, padding=padding, truncation=True)

        # Tokenize targets
        labels = tokenizer(text_target=sample["completion"], max_length=max_target_length, padding=padding, truncation=True)
        labels = labels["input_ids"]

        # Replace all tokenizer.pad_token_id in the labels by IGNORE_INDEX when we want to ignore padding in the loss.
        if padding == "max_length":
            labels = [[(l if l != tokenizer.pad_token_id else IGNORE_INDEX) for l in label] for label in labels]

        model_inputs["labels"] = labels
        return model_inputs

    def get_tokens_causal(sample, max_length, padding="max_length"):
        # Tokenize inputs
        model_inputs = tokenizer(sample["combined"], max_length=max_length, padding=padding, truncation=True)

        # Create targets
        labels = copy.deepcopy(model_inputs["input_ids"])
        prompts = tokenizer(sample["prompt"], max_length=max_length, truncation=True)
        prompt_lens = [len(tokens) for tokens in prompts["input_ids"]]
        for label, source_len in zip(labels, prompt_lens):
            label[:source_len] = [IGNORE_INDEX] * source_len

        # Replace all tokenizer.pad_token_id in the labels by IGNORE_INDEX when we want to ignore padding in the loss.
        if padding == "max_length":
            labels = [[(l if l != tokenizer.pad_token_id else IGNORE_INDEX) for l in label] for label in labels]

        model_inputs["labels"] = labels
        return model_inputs

    if encoder_decoder_model:
        max_source_length = get_dataset_stats(dataset, tokenizer, "prompt")
        max_target_length = get_dataset_stats(dataset, tokenizer, "completion")
        kwargs = {"max_source_length" : max_source_length, "max_target_length" : max_target_length}
        tokenized_dataset = dataset.map(get_tokens_seq2seq, batched=True, fn_kwargs=kwargs)

    elif decoder_only_model:
        dataset = dataset.map(lambda example: {"combined": example["prompt"] + " " + example["completion"]})
        dataset = dataset.filter(lambda x: len(tokenizer(x["combined"])["input_ids"]) <= tokenizer.model_max_length)
        max_length = get_dataset_stats(dataset, tokenizer, "combined")
        kwargs = {"max_length" : max_length}
        tokenized_dataset = dataset.map(get_tokens_causal, batched=True, fn_kwargs=kwargs)

    print(f"Dataset statistics: {kwargs}")
    print(f"Keys of tokenized dataset: {list(tokenized_dataset.features)}")
    return tokenized_dataset


def _compute_metrics(metric, eval_preds, tokenizer):
    preds, labels = eval_preds
    if isinstance(preds, tuple):
        preds = preds[0]
    decoded_preds = tokenizer.batch_decode(preds, skip_special_tokens=True)

    # Replace IGNORE_INDEX in the labels as we can't decode them.
    labels = np.where(labels != IGNORE_INDEX, labels, tokenizer.pad_token_id)
    decoded_labels = tokenizer.batch_decode(labels, skip_special_tokens=True)

    result = metric.compute(predictions=decoded_preds, references=decoded_labels, use_stemmer=True)
    result = {k: round(v * 100, 4) for k, v in result.items()}
    prediction_lens = [np.count_nonzero(pred != tokenizer.pad_token_id) for pred in preds]
    result["gen_len"] = np.mean(prediction_lens)
    return result


class PeftSavingCallback(TrainerCallback):
    def on_train_end(self, args, state, control, **kwargs):
        peft_model_path = state.best_model_checkpoint
        kwargs["model"].save_pretrained(peft_model_path)

        pytorch_model_path = os.path.join(state.best_model_checkpoint, "pytorch_model.bin")
        os.remove(pytorch_model_path) if os.path.exists(pytorch_model_path) else None


def _train_seq2seq(model, tokenizer, tokenized_dataset, metric, config):
    # Define data collator
    data_collator = DataCollatorForSeq2Seq(tokenizer=tokenizer, model=model)

    # Define training args
    training_args = Seq2SeqTrainingArguments(
        output_dir=config['output_dir'],
        per_device_train_batch_size=config['batch_size'],
        gradient_accumulation_steps=config['gradient_accumulation_steps'],
        per_device_eval_batch_size=config['batch_size'],
        predict_with_generate=True,
        learning_rate=config['lr'], #1e-4, # 5e-5
        num_train_epochs=config['epochs'],
        # logging & evaluation strategies
        log_level="error",
        logging_dir=f"{config['output_dir']}/logs",
        logging_strategy="steps",
        logging_steps=500,
        evaluation_strategy="epoch",
        save_strategy="epoch",
        save_total_limit=config['epochs'],
        load_best_model_at_end=True,
        report_to="tensorboard",
        fp16=config['fp16'],
        bf16=config['bf16'],
    )

    # Create trainer instance
    trainer = Seq2SeqTrainer(
        model=model,
        tokenizer=tokenizer,
        args=training_args,
        train_dataset=tokenized_dataset["train"],
        eval_dataset=tokenized_dataset["test"],
        data_collator=data_collator,
        compute_metrics=lambda x: _compute_metrics(metric, x, tokenizer),
        callbacks=[PeftSavingCallback] if config['peft'] else None,
    )

    trainer.train()

    return trainer.state.best_model_checkpoint


def smart_tokenizer_and_embedding_resize(special_tokens_dict, tokenizer, model):
    """
    Resize tokenizer and embedding.
    Note: This is the unoptimized version that may make your embedding size not be divisible by 64.
    """
    num_new_tokens = tokenizer.add_special_tokens(special_tokens_dict)
    model.resize_token_embeddings(len(tokenizer))

    if num_new_tokens > 0:
        input_embeddings = model.get_input_embeddings().weight.data
        output_embeddings = model.get_output_embeddings().weight.data

        input_embeddings_avg = input_embeddings[:-num_new_tokens].mean(dim=0, keepdim=True)
        output_embeddings_avg = output_embeddings[:-num_new_tokens].mean(dim=0, keepdim=True)

        input_embeddings[-num_new_tokens:] = input_embeddings_avg
        output_embeddings[-num_new_tokens:] = output_embeddings_avg


@dataclass
class DataCollatorForSupervisedDataset:
    """
    Collate examples for supervised fine-tuning.
    """
    tokenizer: PreTrainedTokenizer

    def __call__(self, instances):
        pad_token_id = self.tokenizer.pad_token_id

        input_ids, labels = tuple([instance[key] for instance in instances] for key in ("input_ids", "labels"))
        input_ids, labels = torch.tensor(input_ids), torch.tensor(labels)

        input_ids = torch.nn.utils.rnn.pad_sequence(input_ids, batch_first=True, padding_value=pad_token_id)

        labels = torch.nn.utils.rnn.pad_sequence(labels, batch_first=True, padding_value=IGNORE_INDEX)
        return dict(input_ids=input_ids, labels=labels, attention_mask=input_ids.ne(pad_token_id))


def _train_causal(model, tokenizer, tokenized_dataset, metric, config):
    # Define data collator
    data_collator = DataCollatorForSupervisedDataset(tokenizer=tokenizer)

    # Define training args
    training_args = TrainingArguments(
        output_dir=config['output_dir'],
        per_device_train_batch_size=config['batch_size'],
        gradient_accumulation_steps=config['gradient_accumulation_steps'],
        per_device_eval_batch_size=config['batch_size'],
        learning_rate=config['lr'], #1e-4,# 5e-5
        num_train_epochs=config['epochs'],
        # logging & evaluation strategies
        log_level="error",
        logging_dir=f"{config['output_dir']}/logs",
        logging_strategy="steps",
        logging_steps=500,
        evaluation_strategy="epoch",
        save_strategy="epoch",
        save_total_limit=config['epochs'],
        load_best_model_at_end=True,
        report_to="tensorboard",
        fp16=config['fp16'],
        bf16=config['bf16'],
    )

    # Create trainer instance
    trainer = Trainer(
        model=model,
        tokenizer=tokenizer,
        args=training_args,
        train_dataset=tokenized_dataset["train"],
        eval_dataset=tokenized_dataset["test"],
        data_collator=data_collator,
        callbacks=[PeftSavingCallback] if config['peft'] else None,
    )

    trainer.train()

    return trainer.state.best_model_checkpoint


def finetune_hf(data_path, target, config):
    set_seed(42)

    output_dir = os.path.join('../finetuning_ckpts', config['save'])

    if os.path.exists(output_dir):
        # training completed, load best model
        ckpts = glob.glob(f'{output_dir}/checkpoint*')
        final_ckpt = sorted(ckpts, key=lambda x: int(x.split('-')[-1]))[-1]
        with open(os.path.join(final_ckpt, 'trainer_state.json')) as f:
            state = json.load(f)
        best_model_checkpoint = state['best_model_checkpoint']

    else:
        os.makedirs(output_dir, exist_ok=True)
        config['target'] = target
        config['output_dir'] = output_dir
        with open(os.path.join(config['output_dir'], 'compiler_config.json'), 'w') as f:
            json.dump(config, f)

        architecture = AutoConfig.from_pretrained(target).__dict__["architectures"][0]
        encoder_decoder_model = ("ConditionalGeneration" in architecture) or ("T5WithLMHeadModel" in architecture)
        decoder_only_model = ("CausalLM" in architecture) or ("GPT2LMHeadModel" in architecture)
        assert encoder_decoder_model or decoder_only_model, f"Unknown HuggingFace model class: {target}"
        assert not config['fid'] or encoder_decoder_model, "Model must be encoder-decoder for Fusion in Decoder"
        assert not config['fid'] or not config['peft'], "FiD and PEFT can't be trained together"

        # load model
        AutoModelClass = AutoModelForSeq2SeqLM if encoder_decoder_model else AutoModelForCausalLM
        if config['peft']:
            model = AutoModelClass.from_pretrained(target, device_map='auto')
            task_type = TaskType.SEQ_2_SEQ_LM if encoder_decoder_model else TaskType.CAUSAL_LM
            peft_config = LoraConfig(task_type=task_type, inference_mode=False, r=8, lora_alpha=32, lora_dropout=0.1)
            model = get_peft_model(model, peft_config)
            model.print_trainable_parameters()
        else:
            if config['fid']:
                t5 = AutoModelClass.from_pretrained(target)
                model = FiDT5(t5.config)
                model.load_t5(t5.state_dict())
            else:
                model = AutoModelClass.from_pretrained(target)
                # model = _freeze_model_layers(model, unfreeze_last_n=2)

        # load tokenizer
        tokenizer = AutoTokenizer.from_pretrained(target)
        if decoder_only_model:
            smart_tokenizer_and_embedding_resize(SPECIAL_TOKENS_DICT, tokenizer, model)

        # load data
        dataset = _load_data(data_path)
        dataset = _preprocess_data(dataset, tokenizer, encoder_decoder_model, decoder_only_model, config)
        tokenized_dataset = _tokenize_dataset(dataset, tokenizer, encoder_decoder_model, decoder_only_model)
        tokenized_dataset = tokenized_dataset.train_test_split(test_size=0.1)
        print(f'Finetuning dataset: {tokenized_dataset}')

        # start training
        metric = evaluate.load("rouge")
        if encoder_decoder_model:
            best_model_checkpoint = _train_seq2seq(model, tokenizer, tokenized_dataset, metric, config)
        elif decoder_only_model:
            best_model_checkpoint = _train_causal(model, tokenizer, tokenized_dataset, metric, config)

    print(f'Best checkpoint of model: {best_model_checkpoint}')
    return best_model_checkpoint



from .copro_optimizer import COPRO

"""
===============================================================
DEPRECATED!!!
PLEASE USE COPRO INSTEAD.
===============================================================

USAGE SUGGESTIONS:

The following code can be used to compile a optimized signature teleprompter, and evaluate it on an end task:

teleprompter = SignatureOptimizer(prompt_model=prompt_model, metric=metric, breadth=BREADTH, depth=DEPTH, init_temperature=INIT_TEMPERATURE)
kwargs = dict(num_threads=NUM_THREADS, display_progress=True, display_table=0)
compiled_prompt_opt = teleprompter.compile(program.deepcopy(), devset=devset[:DEV_NUM], eval_kwargs=kwargs)
eval_score = evaluate(compiled_prompt_opt, devset=evalset[:EVAL_NUM], **kwargs)

Note that this teleprompter takes in the following parameters:

* prompt_model: The model used for prompt generation. When unspecified, defaults to the model set in settings (ie. dspy.settings.configure(lm=task_model)).
* metric: The task metric used for optimization.
* breadth: The number of new prompts to generate at each iteration. Default=10.
* depth: The number of times we should ask our prompt model to generate new prompts, with the history of the past prompts as input. Default=3.
* init_temperature: The temperature used to generate new prompts. Higher roughly equals more creative. Default=1.4.
* verbose: Tells the method whether or not to print intermediate steps.
* track_stats: Tells the method whether or not to track statistics about the optimization process.
                If True, the method will track the following statistics:
                    * results_best: The min,max,avg,stddev of top 10 scores for each predictor at each depth.
                    * results_latest: The min,max,avg,stddev of newest prompt scores for each predictor at each depth.
                    * total_calls: The total number of calls to the task metric.
                These statistics will be returned as attributes of the best program.
"""


class SignatureOptimizer(COPRO):
    def __init__(
        self,
        prompt_model=None,
        metric=None,
        breadth=10,
        depth=3,
        init_temperature=1.4,
        verbose=False,
        track_stats=False,
    ):
        print(
            "\u001b[31m[WARNING] SignatureOptimizer has been deprecated and replaced with COPRO.  SignatureOptimizer will be removed in a future release. \u001b[31m",
        )
        super().__init__(prompt_model, metric, breadth, depth, init_temperature, verbose, track_stats)

    def compile(self, student, *, devset, eval_kwargs):
        return super().compile(student, trainset=devset, eval_kwargs=eval_kwargs)



import random

from dspy.datasets.dataset import Dataset

### A bunch of colors, originally from matplotlib
all_colors = ['alice blue', 'dodger blue', 'light sky blue', 'deep sky blue', 'sky blue', 'steel blue', 'light steel blue', 'medium blue', 'navy blue', 'blue', 'royal blue', 'cadet blue', 'cornflower blue', 'medium slate blue', 'slate blue', 'dark slate blue', 'powder blue', 'turquoise', 'dark turquoise', 'medium turquoise', 'pale turquoise', 'light sea green', 'medium sea green', 'sea green', 'forest green', 'green yellow', 'lime green', 'dark green', 'green', 'lime', 'chartreuse', 'lawn green', 'yellow green', 'olive green', 'dark olive green', 'medium spring green', 'spring green', 'medium aquamarine', 'aquamarine', 'aqua', 'cyan', 'dark cyan', 'teal', 'medium orchid', 'dark orchid', 'orchid', 'blue violet', 'violet', 'dark violet', 'plum', 'thistle', 'magenta', 'fuchsia', 'dark magenta', 'medium purple', 'purple', 'rebecca purple', 'dark red', 'fire brick', 'indian red', 'light coral', 'dark salmon', 'light salmon', 'salmon', 'red', 'crimson', 'tomato', 'coral', 'orange red', 'dark orange', 'orange', 'yellow', 'gold', 'light goldenrod yellow', 'pale goldenrod', 'goldenrod', 'dark goldenrod', 'beige', 'moccasin', 'blanched almond', 'navajo white', 'antique white', 'bisque', 'burlywood', 'dark khaki', 'khaki', 'tan', 'wheat', 'snow', 'floral white', 'old lace', 'ivory', 'linen', 'seashell', 'honeydew', 'mint cream', 'azure', 'lavender', 'ghost white', 'white smoke', 'gainsboro', 'light gray', 'silver', 'dark gray', 'gray', 'dim gray', 'slate gray', 'light slate gray', 'dark slate gray', 'black', 'medium violet red', 'pale violet red', 'deep pink', 'hot pink', 'light pink', 'pink', 'peach puff', 'rosy brown', 'saddle brown', 'sandy brown', 'chocolate', 'peru', 'sienna', 'brown', 'maroon', 'white', 'misty rose', 'lavender blush', 'papaya whip', 'lemon chiffon', 'light yellow', 'corn silk', 'pale green', 'light green', 'olive drab', 'olive', 'dark sea green']

class Colors(Dataset):
    def __init__(self, sort_by_suffix=True, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.sort_by_suffix = sort_by_suffix
        colors = self.sorted_by_suffix(all_colors)

        train_size = int(len(colors) * 0.6) # chosen to ensure that similar colors aren't repeated between train and dev
        train_colors, dev_colors = colors[:train_size], colors[train_size:]

        self._train = [dict(color=color) for color in train_colors]
        self._dev = [dict(color=color) for color in dev_colors]

        random.Random(0).shuffle(self._train)
        random.Random(0).shuffle(self._dev)
    
    def sorted_by_suffix(self, colors):
        if not self.sort_by_suffix:
            return colors

        if isinstance(colors[0], str):
            sorted_colors = sorted(colors, key=lambda x: x[::-1])
        else:
            sorted_colors = sorted(colors, key=lambda x: x['color'][::-1])

        return sorted_colors
